@@ -189,9 +189,15 @@ notmuch_rb_filenames_each(VALUE self);
 VALUE
 notmuch_rb_query_destroy(VALUE self);
 
+VALUE
+notmuch_rb_query_get_sort(VALUE self);
+
 VALUE
 notmuch_rb_query_set_sort(VALUE self, VALUE sortv);
 
+VALUE
+notmuch_rb_query_get_string(VALUE self);
+
 VALUE
 notmuch_rb_query_search_threads(VALUE self);
 
@@ -262,6 +268,9 @@ notmuch_rb_message_get_replies(VALUE self);
 VALUE
 notmuch_rb_message_get_filename(VALUE self);
 
+VALUE
+notmuch_rb_message_get_filenames(VALUE self);
+
 VALUE
 notmuch_rb_message_get_flag(VALUE self, VALUE flagv);
 
@@ -286,6 +295,12 @@ notmuch_rb_message_remove_tag(VALUE self, VALUE tagv);
 VALUE
 notmuch_rb_message_remove_all_tags(VALUE self);
 
+VALUE
+notmuch_rb_message_maildir_flags_to_tags(VALUE self);
+
+VALUE
+notmuch_rb_message_tags_to_maildir_flags(VALUE self);
+
 VALUE
 notmuch_rb_message_freeze(VALUE self);
 
