@@ -20,25 +20,6 @@
 
 #include "notmuch-client.h"
 
-typedef struct show_format {
-    const char *message_set_start;
-    const char *message_start;
-    void (*message) (const void *ctx,
-		     notmuch_message_t *message,
-		     int indent);
-    const char *header_start;
-    void (*header) (const void *ctx,
-		    notmuch_message_t *message);
-    const char *header_end;
-    const char *body_start;
-    void (*part) (GMimeObject *part,
-		  int *part_count);
-    const char *body_end;
-    const char *message_end;
-    const char *message_set_sep;
-    const char *message_set_end;
-} show_format_t;
-
 static void
 format_message_text (unused (const void *ctx),
 		     notmuch_message_t *message,
@@ -46,14 +27,29 @@ format_message_text (unused (const void *ctx),
 static void
 format_headers_text (const void *ctx,
 		     notmuch_message_t *message);
+
 static void
-format_part_text (GMimeObject *part,
-		  int *part_count);
-static const show_format_t format_text = {
+format_part_start_text (GMimeObject *part,
+			int *part_count);
+
+static void
+format_part_content_text (GMimeObject *part);
+
+static void
+format_part_end_text (GMimeObject *part);
+
+static const notmuch_show_format_t format_text = {
     "",
 	"\fmessage{ ", format_message_text,
 	    "\fheader{\n", format_headers_text, "\fheader}\n",
-	    "\fbody{\n", format_part_text, "\fbody}\n",
+	    "\fbody{\n",
+	        format_part_start_text,
+	        NULL,
+	        NULL,
+	        format_part_content_text,
+	        format_part_end_text,
+	        "",
+	    "\fbody}\n",
 	"\fmessage}\n", "",
     ""
 };
@@ -65,14 +61,35 @@ format_message_json (const void *ctx,
 static void
 format_headers_json (const void *ctx,
 		     notmuch_message_t *message);
+
 static void
-format_part_json (GMimeObject *part,
-		  int *part_count);
-static const show_format_t format_json = {
+format_part_start_json (unused (GMimeObject *part),
+			int *part_count);
+
+static void
+format_part_encstatus_json (int status);
+
+static void
+format_part_sigstatus_json (const GMimeSignatureValidity* validity);
+
+static void
+format_part_content_json (GMimeObject *part);
+
+static void
+format_part_end_json (GMimeObject *part);
+
+static const notmuch_show_format_t format_json = {
     "[",
 	"{", format_message_json,
 	    ", \"headers\": {", format_headers_json, "}",
-	    ", \"body\": [", format_part_json, "]",
+	    ", \"body\": [",
+	        format_part_start_json,
+	        format_part_encstatus_json,
+	        format_part_sigstatus_json,
+	        format_part_content_json,
+	        format_part_end_json,
+	        ", ",
+	    "]",
 	"}", ", ",
     "]"
 };
@@ -82,15 +99,41 @@ format_message_mbox (const void *ctx,
 		     notmuch_message_t *message,
 		     unused (int indent));
 
-static const show_format_t format_mbox = {
+static const notmuch_show_format_t format_mbox = {
     "",
         "", format_message_mbox,
             "", NULL, "",
-            "", NULL, "",
+            "",
+                NULL,
+                NULL,
+                NULL,
+                NULL,
+                NULL,
+                "",
+            "",
         "", "",
     ""
 };
 
+static void
+format_part_content_raw (GMimeObject *part);
+
+static const notmuch_show_format_t format_raw = {
+    "",
+	"", NULL,
+	    "", NULL, "",
+            "",
+                NULL,
+                NULL,
+                NULL,
+                format_part_content_raw,
+                NULL,
+                "",
+            "",
+	"", "",
+    ""
+};
+
 static const char *
 _get_tags_as_string (const void *ctx, notmuch_message_t *message)
 {
@@ -331,29 +374,42 @@ format_headers_json (const void *ctx, notmuch_message_t *message)
     talloc_free (ctx_quote);
 }
 
+/* Write a MIME text part out to the given stream.
+ *
+ * Both line-ending conversion (CRLF->LF) and charset conversion ( ->
+ * UTF-8) will be performed, so it is inappropriate to call this
+ * function with a non-text part. Doing so will trigger an internal
+ * error.
+ */
 static void
-show_part_content (GMimeObject *part, GMimeStream *stream_out)
+show_text_part_content (GMimeObject *part, GMimeStream *stream_out)
 {
+    GMimeContentType *content_type = g_mime_object_get_content_type (GMIME_OBJECT (part));
     GMimeStream *stream_filter = NULL;
     GMimeDataWrapper *wrapper;
     const char *charset;
 
-    charset = g_mime_object_get_content_type_parameter (part, "charset");
+    if (! g_mime_content_type_is_type (content_type, "text", "*"))
+	INTERNAL_ERROR ("Illegal request to format non-text part (%s) as text.",
+			g_mime_content_type_to_string (content_type));
 
-    if (stream_out) {
-	stream_filter = g_mime_stream_filter_new (stream_out);
-	g_mime_stream_filter_add(GMIME_STREAM_FILTER (stream_filter),
-				 g_mime_filter_crlf_new (FALSE, FALSE));
-        if (charset) {
-	    GMimeFilter *charset_filter;
-	    charset_filter = g_mime_filter_charset_new (charset, "UTF-8");
-	    /* This result can be NULL for things like "unknown-8bit".
-	     * Don't set a NULL filter as that makes GMime print
-	     * annoying assertion-failure messages on stderr. */
-	    if (charset_filter)
-		g_mime_stream_filter_add (GMIME_STREAM_FILTER (stream_filter),
-					  charset_filter);
-	}
+    if (stream_out == NULL)
+	return;
+
+    stream_filter = g_mime_stream_filter_new (stream_out);
+    g_mime_stream_filter_add(GMIME_STREAM_FILTER (stream_filter),
+			     g_mime_filter_crlf_new (FALSE, FALSE));
+
+    charset = g_mime_object_get_content_type_parameter (part, "charset");
+    if (charset) {
+	GMimeFilter *charset_filter;
+	charset_filter = g_mime_filter_charset_new (charset, "UTF-8");
+	/* This result can be NULL for things like "unknown-8bit".
+	 * Don't set a NULL filter as that makes GMime print
+	 * annoying assertion-failure messages on stderr. */
+	if (charset_filter)
+	    g_mime_stream_filter_add (GMIME_STREAM_FILTER (stream_filter),
+				      charset_filter);
     }
 
     wrapper = g_mime_part_get_content_object (GMIME_PART (part));
@@ -363,81 +419,187 @@ show_part_content (GMimeObject *part, GMimeStream *stream_out)
 	g_object_unref(stream_filter);
 }
 
-static void
-format_part_text (GMimeObject *part, int *part_count)
+static const char*
+signer_status_to_string (GMimeSignerStatus x)
 {
-    GMimeContentDisposition *disposition;
-    GMimeContentType *content_type;
+    switch (x) {
+    case GMIME_SIGNER_STATUS_NONE:
+	return "none";
+    case GMIME_SIGNER_STATUS_GOOD:
+	return "good";
+    case GMIME_SIGNER_STATUS_BAD:
+	return "bad";
+    case GMIME_SIGNER_STATUS_ERROR:
+	return "error";
+    }
+    return "unknown";
+}
+
+static void
+format_part_start_text (GMimeObject *part, int *part_count)
+{
+    GMimeContentDisposition *disposition = g_mime_object_get_content_disposition (part);
+
+    if (disposition &&
+	strcmp (disposition->disposition, GMIME_DISPOSITION_ATTACHMENT) == 0)
+    {
+	printf ("\fattachment{ ID: %d", *part_count);
+
+    } else {
+
+	printf ("\fpart{ ID: %d", *part_count);
+    }
+}
+
+static void
+format_part_content_text (GMimeObject *part)
+{
+    GMimeContentDisposition *disposition = g_mime_object_get_content_disposition (part);
+    GMimeContentType *content_type = g_mime_object_get_content_type (GMIME_OBJECT (part));
+
+    printf (", Content-type: %s\n", g_mime_content_type_to_string (content_type));
 
-    disposition = g_mime_object_get_content_disposition (part);
     if (disposition &&
 	strcmp (disposition->disposition, GMIME_DISPOSITION_ATTACHMENT) == 0)
     {
 	const char *filename = g_mime_part_get_filename (GMIME_PART (part));
-	content_type = g_mime_object_get_content_type (GMIME_OBJECT (part));
-
-	printf ("\fattachment{ ID: %d, Content-type: %s\n",
-		*part_count,
-		g_mime_content_type_to_string (content_type));
 	printf ("Attachment: %s (%s)\n", filename,
 		g_mime_content_type_to_string (content_type));
-
-	if (g_mime_content_type_is_type (content_type, "text", "*") &&
-	    !g_mime_content_type_is_type (content_type, "text", "html"))
-	{
-	    GMimeStream *stream_stdout = g_mime_stream_file_new (stdout);
-	    g_mime_stream_file_set_owner (GMIME_STREAM_FILE (stream_stdout), FALSE);
-	    show_part_content (part, stream_stdout);
-	    g_object_unref(stream_stdout);
-	}
-
-	printf ("\fattachment}\n");
-
-	return;
     }
 
-    content_type = g_mime_object_get_content_type (GMIME_OBJECT (part));
-
-    printf ("\fpart{ ID: %d, Content-type: %s\n",
-	    *part_count,
-	    g_mime_content_type_to_string (content_type));
-
     if (g_mime_content_type_is_type (content_type, "text", "*") &&
 	!g_mime_content_type_is_type (content_type, "text", "html"))
     {
 	GMimeStream *stream_stdout = g_mime_stream_file_new (stdout);
 	g_mime_stream_file_set_owner (GMIME_STREAM_FILE (stream_stdout), FALSE);
-	show_part_content (part, stream_stdout);
+	show_text_part_content (part, stream_stdout);
 	g_object_unref(stream_stdout);
     }
+    else if (g_mime_content_type_is_type (content_type, "multipart", "*") ||
+	     g_mime_content_type_is_type (content_type, "message", "rfc822"))
+    {
+	/* Do nothing for multipart since its content will be printed
+	 * when recursing. */
+    }
     else
     {
 	printf ("Non-text part: %s\n",
 		g_mime_content_type_to_string (content_type));
     }
-
-    printf ("\fpart}\n");
 }
 
 static void
-format_part_json (GMimeObject *part, int *part_count)
+format_part_end_text (GMimeObject *part)
 {
-    GMimeContentType *content_type;
     GMimeContentDisposition *disposition;
-    void *ctx = talloc_new (NULL);
-    GMimeStream *stream_memory = g_mime_stream_mem_new ();
-    GByteArray *part_content;
-
-    content_type = g_mime_object_get_content_type (GMIME_OBJECT (part));
-
-    if (*part_count > 1)
-	fputs (", ", stdout);
-
-    printf ("{\"id\": %d, \"content-type\": %s",
-	    *part_count,
-	    json_quote_str (ctx, g_mime_content_type_to_string (content_type)));
 
     disposition = g_mime_object_get_content_disposition (part);
+    if (disposition &&
+	strcmp (disposition->disposition, GMIME_DISPOSITION_ATTACHMENT) == 0)
+    {
+	printf ("\fattachment}\n");
+    }
+    else
+    {
+	printf ("\fpart}\n");
+    }
+}
+
+static void
+format_part_start_json (unused (GMimeObject *part), int *part_count)
+{
+    printf ("{\"id\": %d", *part_count);
+}
+
+static void
+format_part_encstatus_json (int status)
+{
+    printf (", \"encstatus\": [{\"status\": ");
+    if (status) {
+	printf ("\"good\"");
+    } else {
+	printf ("\"bad\"");
+    }
+    printf ("}]");
+}
+
+static void
+format_part_sigstatus_json (const GMimeSignatureValidity* validity)
+{
+    printf (", \"sigstatus\": [");
+
+    if (!validity) {
+	printf ("]");
+	return;
+    }
+
+    const GMimeSigner *signer = g_mime_signature_validity_get_signers (validity);
+    int first = 1;
+    void *ctx_quote = talloc_new (NULL);
+
+    while (signer) {
+	if (first)
+	    first = 0;
+	else
+	    printf (", ");
+
+	printf ("{");
+
+	/* status */
+	printf ("\"status\": %s",
+		json_quote_str (ctx_quote,
+				signer_status_to_string (signer->status)));
+
+	if (signer->status == GMIME_SIGNER_STATUS_GOOD)
+	{
+	    if (signer->fingerprint)
+		printf (", \"fingerprint\": %s", json_quote_str (ctx_quote, signer->fingerprint));
+	    /* these dates are seconds since the epoch; should we
+	     * provide a more human-readable format string? */
+	    if (signer->created)
+		printf (", \"created\": %d", (int) signer->created);
+	    if (signer->expires)
+		printf (", \"expires\": %d", (int) signer->expires);
+	    /* output user id only if validity is FULL or ULTIMATE. */
+	    /* note that gmime is using the term "trust" here, which
+	     * is WRONG.  It's actually user id "validity". */
+	    if ((signer->name) && (signer->trust)) {
+		if ((signer->trust == GMIME_SIGNER_TRUST_FULLY) || (signer->trust == GMIME_SIGNER_TRUST_ULTIMATE))
+		    printf (", \"userid\": %s", json_quote_str (ctx_quote, signer->name));
+           }
+       } else {
+           if (signer->keyid)
+               printf (", \"keyid\": %s", json_quote_str (ctx_quote, signer->keyid));
+       }
+       if (signer->errors != GMIME_SIGNER_ERROR_NONE) {
+           printf (", \"errors\": %x", signer->errors);
+       }
+
+       printf ("}");
+       signer = signer->next;
+    }
+
+    printf ("]");
+
+    talloc_free (ctx_quote);
+}
+
+static void
+format_part_content_json (GMimeObject *part)
+{
+    GMimeContentType *content_type = g_mime_object_get_content_type (GMIME_OBJECT (part));
+    GMimeStream *stream_memory = g_mime_stream_mem_new ();
+    const char *cid = g_mime_object_get_content_id (part);
+    void *ctx = talloc_new (NULL);
+    GMimeContentDisposition *disposition = g_mime_object_get_content_disposition (part);
+    GByteArray *part_content;
+
+    printf (", \"content-type\": %s",
+	    json_quote_str (ctx, g_mime_content_type_to_string (content_type)));
+
+    if (cid != NULL)
+	    printf(", \"content-id\": %s", json_quote_str (ctx, cid));
+
     if (disposition &&
 	strcmp (disposition->disposition, GMIME_DISPOSITION_ATTACHMENT) == 0)
     {
@@ -449,13 +611,16 @@ format_part_json (GMimeObject *part, int *part_count)
     if (g_mime_content_type_is_type (content_type, "text", "*") &&
 	!g_mime_content_type_is_type (content_type, "text", "html"))
     {
-	show_part_content (part, stream_memory);
+	show_text_part_content (part, stream_memory);
 	part_content = g_mime_stream_mem_get_byte_array (GMIME_STREAM_MEM (stream_memory));
 
 	printf (", \"content\": %s", json_quote_chararray (ctx, (char *) part_content->data, part_content->len));
     }
-
-    fputs ("}", stdout);
+    else if (g_mime_content_type_is_type (content_type, "multipart", "*") ||
+	     g_mime_content_type_is_type (content_type, "message", "rfc822"))
+    {
+	printf (", \"content\": [");
+    }
 
     talloc_free (ctx);
     if (stream_memory)
@@ -463,29 +628,80 @@ format_part_json (GMimeObject *part, int *part_count)
 }
 
 static void
-show_message (void *ctx, const show_format_t *format, notmuch_message_t *message, int indent)
+format_part_end_json (GMimeObject *part)
 {
-    fputs (format->message_start, stdout);
-    if (format->message)
-	format->message(ctx, message, indent);
+    GMimeContentType *content_type;
 
-    fputs (format->header_start, stdout);
-    if (format->header)
-	format->header(ctx, message);
-    fputs (format->header_end, stdout);
+    content_type = g_mime_object_get_content_type (GMIME_OBJECT (part));
 
-    fputs (format->body_start, stdout);
-    if (format->part)
-	show_message_body (notmuch_message_get_filename (message), format->part);
-    fputs (format->body_end, stdout);
+    if (g_mime_content_type_is_type (content_type, "multipart", "*") ||
+	g_mime_content_type_is_type (content_type, "message", "rfc822"))
+	printf ("]");
 
-    fputs (format->message_end, stdout);
+    printf ("}");
 }
 
+static void
+format_part_content_raw (GMimeObject *part)
+{
+    GMimeStream *stream_stdout;
+    GMimeStream *stream_filter = NULL;
+    GMimeDataWrapper *wrapper;
+
+    stream_stdout = g_mime_stream_file_new (stdout);
+    g_mime_stream_file_set_owner (GMIME_STREAM_FILE (stream_stdout), FALSE);
+
+    stream_filter = g_mime_stream_filter_new (stream_stdout);
+
+    wrapper = g_mime_part_get_content_object (GMIME_PART (part));
+
+    if (wrapper && stream_filter)
+	g_mime_data_wrapper_write_to_stream (wrapper, stream_filter);
+
+    if (stream_filter)
+	g_object_unref (stream_filter);
+
+    if (stream_stdout)
+	g_object_unref(stream_stdout);
+}
 
 static void
-show_messages (void *ctx, const show_format_t *format, notmuch_messages_t *messages, int indent,
-	       notmuch_bool_t entire_thread)
+show_message (void *ctx,
+	      const notmuch_show_format_t *format,
+	      notmuch_message_t *message,
+	      int indent,
+	      notmuch_show_params_t *params)
+{
+    if (params->part <= 0) {
+	fputs (format->message_start, stdout);
+	if (format->message)
+	    format->message(ctx, message, indent);
+
+	fputs (format->header_start, stdout);
+	if (format->header)
+	    format->header(ctx, message);
+	fputs (format->header_end, stdout);
+
+	fputs (format->body_start, stdout);
+    }
+
+    if (format->part_content)
+	show_message_body (notmuch_message_get_filename (message),
+			   format, params);
+
+    if (params->part <= 0) {
+	fputs (format->body_end, stdout);
+
+	fputs (format->message_end, stdout);
+    }
+}
+
+static void
+show_messages (void *ctx,
+	       const notmuch_show_format_t *format,
+	       notmuch_messages_t *messages,
+	       int indent,
+	       notmuch_show_params_t *params)
 {
     notmuch_message_t *message;
     notmuch_bool_t match;
@@ -510,15 +726,18 @@ show_messages (void *ctx, const show_format_t *format, notmuch_messages_t *messa
 
 	next_indent = indent;
 
-	if (match || entire_thread) {
-	    show_message (ctx, format, message, indent);
+	if (match || params->entire_thread) {
+	    show_message (ctx, format, message, indent, params);
 	    next_indent = indent + 1;
 
 	    fputs (format->message_set_sep, stdout);
 	}
 
-	show_messages (ctx, format, notmuch_message_get_replies (message),
-		       next_indent, entire_thread);
+	show_messages (ctx,
+		       format,
+		       notmuch_message_get_replies (message),
+		       next_indent,
+		       params);
 
 	notmuch_message_destroy (message);
 
@@ -528,16 +747,15 @@ show_messages (void *ctx, const show_format_t *format, notmuch_messages_t *messa
     fputs (format->message_set_end, stdout);
 }
 
-/* Support for --format=raw */
+/* Formatted output of single message */
 static int
-do_show_raw (unused(void *ctx), notmuch_query_t *query)
+do_show_single (void *ctx,
+		notmuch_query_t *query,
+		const notmuch_show_format_t *format,
+		notmuch_show_params_t *params)
 {
     notmuch_messages_t *messages;
     notmuch_message_t *message;
-    const char *filename;
-    FILE *file;
-    size_t size;
-    char buf[4096];
 
     if (notmuch_query_count_messages (query) != 1) {
 	fprintf (stderr, "Error: search term did not match precisely one message.\n");
@@ -552,34 +770,50 @@ do_show_raw (unused(void *ctx), notmuch_query_t *query)
 	return 1;
     }
 
-    filename = notmuch_message_get_filename (message);
-    if (filename == NULL) {
-	fprintf (stderr, "Error: Cannot message filename.\n");
-	return 1;
-    }
+    notmuch_message_set_flag (message, NOTMUCH_MESSAGE_FLAG_MATCH, 1);
 
-    file = fopen (filename, "r");
-    if (file == NULL) {
-	fprintf (stderr, "Error: Cannot open file %s: %s\n", filename, strerror (errno));
-	return 1;
-    }
+    /* Special case for --format=raw of full single message, just cat out file */
+    if (params->raw && 0 == params->part) {
 
-    while (!feof (file)) {
-	size = fread (buf, 1, sizeof (buf), file);
-	fwrite (buf, size, 1, stdout);
-    }
+	const char *filename;
+	FILE *file;
+	size_t size;
+	char buf[4096];
 
-    fclose (file);
+	filename = notmuch_message_get_filename (message);
+	if (filename == NULL) {
+	    fprintf (stderr, "Error: Cannot message filename.\n");
+	    return 1;
+	}
+
+	file = fopen (filename, "r");
+	if (file == NULL) {
+	    fprintf (stderr, "Error: Cannot open file %s: %s\n", filename, strerror (errno));
+	    return 1;
+	}
+
+	while (!feof (file)) {
+	    size = fread (buf, 1, sizeof (buf), file);
+	    fwrite (buf, size, 1, stdout);
+	}
+
+	fclose (file);
+
+    } else {
+
+	show_message (ctx, format, message, 0, params);
+
+    }
 
     return 0;
 }
 
-/* Support for --format=text|json|mbox */
+/* Formatted output of threads */
 static int
 do_show (void *ctx,
 	 notmuch_query_t *query,
-	 const show_format_t *format,
-	 int entire_thread)
+	 const notmuch_show_format_t *format,
+	 notmuch_show_params_t *params)
 {
     notmuch_threads_t *threads;
     notmuch_thread_t *thread;
@@ -604,7 +838,7 @@ do_show (void *ctx,
 	    fputs (format->message_set_sep, stdout);
 	first_toplevel = 0;
 
-	show_messages (ctx, format, messages, 0, entire_thread);
+	show_messages (ctx, format, messages, 0, params);
 
 	notmuch_thread_destroy (thread);
 
@@ -623,10 +857,17 @@ notmuch_show_command (void *ctx, unused (int argc), unused (char *argv[]))
     notmuch_query_t *query;
     char *query_string;
     char *opt;
-    const show_format_t *format = &format_text;
-    int entire_thread = 0;
+    const notmuch_show_format_t *format = &format_text;
+    notmuch_show_params_t params;
+    int mbox = 0;
+    int format_specified = 0;
     int i;
-    int raw = 0;
+
+    params.entire_thread = 0;
+    params.raw = 0;
+    params.part = -1;
+    params.cryptoctx = NULL;
+    params.decrypt = 0;
 
     for (i = 0; i < argc && argv[i][0] == '-'; i++) {
 	if (strcmp (argv[i], "--") == 0) {
@@ -639,17 +880,35 @@ notmuch_show_command (void *ctx, unused (int argc), unused (char *argv[]))
 		format = &format_text;
 	    } else if (strcmp (opt, "json") == 0) {
 		format = &format_json;
-		entire_thread = 1;
+		params.entire_thread = 1;
 	    } else if (strcmp (opt, "mbox") == 0) {
 		format = &format_mbox;
+		mbox = 1;
 	    } else if (strcmp (opt, "raw") == 0) {
-		raw = 1;
+		format = &format_raw;
+		params.raw = 1;
 	    } else {
 		fprintf (stderr, "Invalid value for --format: %s\n", opt);
 		return 1;
 	    }
+	    format_specified = 1;
+	} else if (STRNCMP_LITERAL (argv[i], "--part=") == 0) {
+	    params.part = atoi(argv[i] + sizeof ("--part=") - 1);
 	} else if (STRNCMP_LITERAL (argv[i], "--entire-thread") == 0) {
-	    entire_thread = 1;
+	    params.entire_thread = 1;
+	} else if ((STRNCMP_LITERAL (argv[i], "--verify") == 0) ||
+		   (STRNCMP_LITERAL (argv[i], "--decrypt") == 0)) {
+	    if (params.cryptoctx == NULL) {
+		GMimeSession* session = g_object_new(g_mime_session_get_type(), NULL);
+		if (NULL == (params.cryptoctx = g_mime_gpg_context_new(session, "gpg")))
+		    fprintf (stderr, "Failed to construct gpg context.\n");
+		else
+		    g_mime_gpg_context_set_always_trust((GMimeGpgContext*)params.cryptoctx, FALSE);
+		g_object_unref (session);
+		session = NULL;
+	    }
+	    if (STRNCMP_LITERAL (argv[i], "--decrypt") == 0)
+		params.decrypt = 1;
 	} else {
 	    fprintf (stderr, "Unrecognized option: %s\n", argv[i]);
 	    return 1;
@@ -669,6 +928,11 @@ notmuch_show_command (void *ctx, unused (int argc), unused (char *argv[]))
 	return 1;
     }
 
+    if (mbox && params.part > 0) {
+	fprintf (stderr, "Error: specifying parts is incompatible with mbox output format.\n");
+	return 1;
+    }
+
     if (*query_string == '\0') {
 	fprintf (stderr, "Error: notmuch show requires at least one search term.\n");
 	return 1;
@@ -685,88 +949,25 @@ notmuch_show_command (void *ctx, unused (int argc), unused (char *argv[]))
 	return 1;
     }
 
-    if (raw)
-	return do_show_raw (ctx, query);
+    /* if part was requested and format was not specified, use format=raw */
+    if (params.part >= 0 && !format_specified)
+	format = &format_raw;
+
+    /* If --format=raw specified without specifying part, we can only
+     * output single message, so set part=0 */
+    if (params.raw && params.part < 0)
+	params.part = 0;
+
+    if (params.part >= 0)
+	return do_show_single (ctx, query, format, &params);
     else
-	return do_show (ctx, query, format, entire_thread);
+	return do_show (ctx, query, format, &params);
 
     notmuch_query_destroy (query);
     notmuch_database_close (notmuch);
 
+    if (params.cryptoctx)
+	g_object_unref(params.cryptoctx);
+
     return 0;
 }
-
-int
-notmuch_part_command (void *ctx, unused (int argc), unused (char *argv[]))
-{
-	notmuch_config_t *config;
-	notmuch_database_t *notmuch;
-	notmuch_query_t *query;
-	notmuch_messages_t *messages;
-	notmuch_message_t *message;
-	char *query_string;
-	int i;
-	int part = 0;
-
-	for (i = 0; i < argc && argv[i][0] == '-'; i++) {
-		if (strcmp (argv[i], "--") == 0) {
-			i++;
-			break;
-		}
-		if (STRNCMP_LITERAL (argv[i], "--part=") == 0) {
-			part = atoi(argv[i] + sizeof ("--part=") - 1);
-		} else {
-			fprintf (stderr, "Unrecognized option: %s\n", argv[i]);
-			return 1;
-		}
-	}
-
-	argc -= i;
-	argv += i;
-
-	config = notmuch_config_open (ctx, NULL, NULL);
-	if (config == NULL)
-		return 1;
-
-	query_string = query_string_from_args (ctx, argc, argv);
-	if (query_string == NULL) {
-		fprintf (stderr, "Out of memory\n");
-		return 1;
-	}
-
-	if (*query_string == '\0') {
-		fprintf (stderr, "Error: notmuch part requires at least one search term.\n");
-		return 1;
-	}
-
-	notmuch = notmuch_database_open (notmuch_config_get_database_path (config),
-					 NOTMUCH_DATABASE_MODE_READ_ONLY);
-	if (notmuch == NULL)
-		return 1;
-
-	query = notmuch_query_create (notmuch, query_string);
-	if (query == NULL) {
-		fprintf (stderr, "Out of memory\n");
-		return 1;
-	}
-
-	if (notmuch_query_count_messages (query) != 1) {
-		fprintf (stderr, "Error: search term did not match precisely one message.\n");
-		return 1;
-	}
-
-	messages = notmuch_query_search_messages (query);
-	message = notmuch_messages_get (messages);
-
-	if (message == NULL) {
-		fprintf (stderr, "Error: cannot find matching message.\n");
-		return 1;
-	}
-
-	show_one_part (notmuch_message_get_filename (message), part);
-
-	notmuch_query_destroy (query);
-	notmuch_database_close (notmuch);
-
-	return 0;
-}
